import { refObject, globalEvents, JSConsole, Vector, world, Dice, Color } from "@tabletop-playground/api";
import { event } from "./promise-utils";

enum TokenKind {
  Literal,
  Symbol
}

interface LiteralToken {
  kind: TokenKind.Literal,
  value: number;
}

interface SymbolToken {
  kind: TokenKind.Symbol,
  symbol: string;
}

type Token = LiteralToken | SymbolToken;

enum ExpressionKind {
  Literal,
  BinOp,
  UnOp
}

interface LiteralExpression {
  kind: ExpressionKind.Literal,
  value: number;
}

interface BinOpExpression {
  kind: ExpressionKind.BinOp,
  op: string,
  lhs: Expression;
  rhs: Expression;
}

interface UnOpExpression {
  kind: ExpressionKind.UnOp,
  op: string,
  rhs: Expression;
}

type Expression = LiteralExpression | BinOpExpression | UnOpExpression;

interface ParseContext {
  tokens: Token[],
  index: number;
}

interface ParseError {
  error: string;
}

interface DiceType {
  template: string;
}

const defaultDiceTypes: Record<number, DiceType> = {
  4: { template: "1885447D4CF808B36797CFB1DD679BAC" },
  6: { template: "A897158B490E36F0911B03B3BE9BA52A" },
  8: { template: "10614E404E82F969E6CFD48CAC80F363" },
  10: { template: "9065AC5141F87F8ADE1F5AB6390BBEE4" },
  12: { template: "9FD625E14B5EEEA9C6C998B2DB3E9085" },
  20: { template: "0A2C628E4A706A123AA3CF9C34CAB9A1" },
}

const BinOps: Record<string, [number, number]> = {
  ["+"]: [6, 6],
  ["-"]: [6, 6],
  ["*"]: [7, 7],
  ["/"]: [7, 7],
  ["%"]: [7, 7],
  ["^"]: [10, 9],
  ["d"]: [11, 11]
}
const UnOpLevel = 8;

const mainObject = refObject

function tokenize(expression: string): Token[] {
  const tokens: Token[] = []
  const length = expression.length;
  let index = 0;
  while (index < length) {
    let next = expression.substring(index, index + 1)
    if (next == " ") {
      index++;
    }
    else if (BinOps[next] || next === "(" || next === ")") {
      index++;
      tokens.push({
        kind: TokenKind.Symbol,
        symbol: next
      })
    }
    else if (!isNaN(Number(next))) {
      const start = index;
      do {
        index++;
        next = expression.substring(index, index + 1);
      } while (index < length && (!isNaN(Number(next)) || next == "."));
      tokens.push({
        kind: TokenKind.Literal,
        value: Number(expression.substring(start, index))
      })
    }
  }
  return tokens;
}

function parseExpression(
  ctx: ParseContext,
  level: number = 0
): Expression | ParseError {
  let tok = ctx.tokens[ctx.index];
  if (!tok) return { error: "Unexpected end of string." };
  let lhs: Expression | ParseError;
  if (tok.kind === TokenKind.Symbol && tok.symbol === "-") {
    ctx.index++;
    const rhs = parseExpression(ctx, UnOpLevel);
    if ("error" in rhs) return rhs;
    lhs = { kind: ExpressionKind.UnOp, op: tok.symbol, rhs }
  }
  else {
    if (tok.kind === TokenKind.Literal) {
      ctx.index++;
      lhs = { kind: ExpressionKind.Literal, value: tok.value };
    }
    else if (tok.symbol === "(") {
      ctx.index++;
      lhs = parseExpression(ctx);
      if ("error" in lhs) return lhs;
      tok = ctx.tokens[ctx.index];
      if (!tok) return { error: "Unexpected end of string." };
      if (tok.kind !== TokenKind.Symbol || tok.symbol !== ")") {
        return { error: "Missing close paren." };
      }
      ctx.index++;
    }
    else {
      return { error: `Unexpected symbol: ${tok.symbol}.` };
    }
  }

  const length = ctx.tokens.length;
  while (ctx.index < length) {
    tok = ctx.tokens[ctx.index];
    if (!tok) return { error: "Unexpected end of string." };
    if (tok.kind !== TokenKind.Symbol) {
      return { error: "Missing symbol or operator." };
    }
    const op = tok.symbol;
    const prio = BinOps[op];
    if (prio && prio[0] > level) {
      ctx.index++;
      const rhs = parseExpression(ctx, prio[1]);
      if ("error" in rhs) return rhs;
      lhs = { kind: ExpressionKind.BinOp, op, lhs, rhs }
    }
    else break;
  }
  return lhs;
}

async function roll(expression: Expression) {
  const dice = spawnDice(getDiceForExpression(expression));
  const promises: Promise<[d: Dice]>[] = Array(dice.length);
  for (const [i, d] of dice.entries()) {
    d.roll();
    promises[i] = event(d.onMovementStopped);
  }
  const values = (await Promise.all(promises)).map(([d]) => Number(d.getCurrentFaceName()));
  const printed = printRolledExpression(expression, values);
  const result = evaluateRolledExpression(expression, values);
  world.broadcastChatMessage(
    `Rolled: ${printed} = ${result}`,
    new Color(0, 1, 1)
  );
}

function getDiceForExpression(expression: Expression): DiceType[] {
  switch (expression.kind) {
    case ExpressionKind.BinOp:
      if (expression.op == "d") {
        if (expression.lhs.kind !== ExpressionKind.Literal) {
          throw `Non-literal dice amount not supported`;
        }
        if (expression.rhs.kind !== ExpressionKind.Literal) {
          throw `Non-literal dice type not supported`;
        }
        if (!(expression.rhs.value in defaultDiceTypes)) {
          throw `Dice type d${expression.rhs.value} not supported`;
        }
        return Array(expression.lhs.value)
          .fill(defaultDiceTypes[expression.rhs.value]);
      }
      else {
        return [
          ...getDiceForExpression(expression.lhs),
          ...getDiceForExpression(expression.rhs)
        ];
      }
    case ExpressionKind.UnOp:
      return getDiceForExpression(expression.rhs);
    case ExpressionKind.Literal:
      return [];
  }
}

function spawnDice(types: DiceType[]): Dice[] {
  const dice: Dice[] = Array(types.length);
  const rollerPos = mainObject.getPosition().add(new Vector(0, 1, 0));
  for (const [i, v] of types.entries()) {
    const d = world.createObjectFromTemplate(v.template, rollerPos);
    if (!(d instanceof Dice)) throw "Dice Expected";
    dice[i] = d;
  }
  return dice;
}

function getExpressionPriority(expression: Expression): number {
  switch (expression.kind) {
    case ExpressionKind.BinOp:
      return BinOps[expression.op][0];
    case ExpressionKind.UnOp:
      return UnOpLevel;
    case ExpressionKind.Literal:
      return Infinity;
  }
}

function printRolledExpression(expression: Expression, values: number[]): string {
  let index = 0;
  function printSubExpression(expression: Expression): string {
    switch (expression.kind) {
      case ExpressionKind.BinOp:
        if (expression.op == "d") {
          if (expression.lhs.kind !== ExpressionKind.Literal) {
            throw `Non-literal dice amount not supported`;
          }
          const segment = values.slice(index, index + expression.lhs.value);
          index += expression.lhs.value;
          const sum = segment.reduce((a, v) => a + v);
          return `[${segment.join("+")}=${sum}]`;
        }
        else {
          const prio = BinOps[expression.op];
          let lhs = printSubExpression(expression.lhs);
          let rhs = printSubExpression(expression.rhs);
          let op = expression.op;
          if (getExpressionPriority(expression.lhs) < prio[0]) {
            lhs = `(${lhs})`;
          }
          if (getExpressionPriority(expression.rhs) <= prio[1]) {
            rhs = `(${rhs})`;
          }
          if (op !== "^") {
            op = ` ${op} `;
          }
          return `${lhs}${op}${rhs}`;
        }
      case ExpressionKind.UnOp:
        let rhs = printSubExpression(expression.rhs);
        if (getExpressionPriority(expression.rhs) < UnOpLevel) {
          rhs = `(${rhs})`;
        }
        return `${expression.op}${expression.rhs}`;
      case ExpressionKind.Literal:
        return String(expression.value);
    }
  }
  return printSubExpression(expression);
}

function evaluateRolledExpression(expression: Expression, values: number[]): number {
  let index = 0;
  function evaluateSubExpression(expression: Expression): number {
    switch (expression.kind) {
      case ExpressionKind.BinOp:
        if (expression.op == "d") {
          if (expression.lhs.kind !== ExpressionKind.Literal) {
            throw `Non-literal dice amount not supported`;
          }
          const segment = values.slice(index, index + expression.lhs.value);
          index += expression.lhs.value;
          // sum
          return segment.reduce((a, v) => a + v);
        }
        else {
          let lhs = evaluateSubExpression(expression.lhs);
          let rhs = evaluateSubExpression(expression.rhs);
          switch (expression.op) {
            case "+": return lhs + rhs;
            case "-": return lhs - rhs;
            case "*": return lhs * rhs;
            case "/": return lhs / rhs;
            case "%": return lhs % rhs;
            case "^": return lhs ^ rhs;
            default: throw "Unrecognised Operation";
          }
        }
      case ExpressionKind.UnOp:
        let rhs = evaluateSubExpression(expression.rhs);
          switch (expression.op) {
            case "-": return -rhs;
            default: throw "Unrecognised Operation";
          }
      case ExpressionKind.Literal:
        return expression.value;
    }
  }
  return evaluateSubExpression(expression);
}

globalEvents.onChatMessage.add(async (sender, message) => {
  if (message.startsWith("/roll")) {
    const expr = parseExpression({
      tokens: tokenize(message.substring(5)),
      index: 0
    });
    JSConsole.log(JSON.stringify(expr));
    if (!("error" in expr)) {
      JSConsole.log("rolling");
      try {
        await roll(expr);
      }
      catch (e: unknown) {
        JSConsole.log(`error: ${e}`);
      }
    }
  }
})
